package com.medm.devicekit.sample

import android.bluetooth.BluetoothAdapter
import android.bluetooth.BluetoothClass.Device
import android.content.Context
import android.content.Intent
import android.os.Bundle
import android.view.Menu
import android.view.MenuItem
import android.widget.ArrayAdapter
import android.widget.AutoCompleteTextView
import android.widget.RadioGroup
import android.widget.Toast
import androidx.activity.result.contract.ActivityResultContracts
import androidx.appcompat.app.AlertDialog
import androidx.appcompat.app.AppCompatActivity
import androidx.core.view.isVisible
import androidx.fragment.app.Fragment
import androidx.fragment.app.FragmentActivity
import androidx.fragment.app.FragmentContainerView
import androidx.fragment.app.ListFragment
import com.google.android.gms.common.config.GservicesValue.isInitialized
import com.google.android.material.bottomnavigation.BottomNavigationView
import com.google.android.material.datepicker.MaterialDatePicker
import com.google.android.material.dialog.MaterialAlertDialogBuilder
import com.google.android.material.progressindicator.LinearProgressIndicator
import com.google.android.material.snackbar.BaseTransientBottomBar
import com.google.android.material.snackbar.Snackbar
import com.google.android.material.switchmaterial.SwitchMaterial
import com.google.android.material.textfield.TextInputEditText
import com.medm.devicekit.*
import org.json.JSONArray
import java.text.SimpleDateFormat
import java.util.*

class MainActivity : AppCompatActivity()
{
    private val LICENSE_SHARED_PREF_ENTRY_NAME = "License"
    private val DEVICE_KIT_LICENSE = "YourLicenseHere" // You can also just pick your license in the app
    private val SELECTED_DEVICES_SHARED_PREF_ENTRY_NAME = "SelectedDevices"
    private lateinit var mSelectedDevices: BooleanArray
    private val mLocationServicesRequester = LocationServicesRequester(this)
    private val mDataCollectionFragment = DataCollectionFragment(mLocationServicesRequester)
    private val mAddedDevicesFragment = AddedDevicesFragment()
    private val mDeviceScanFragment = DeviceScanFragment(mLocationServicesRequester)
    private var mCollectorStopToken: CollectorStopToken? = null
    private var mScanStopToken: ScannerStopToken? = null
    private var mErrorSnackbar: Snackbar? = null
    private val mPickLicense = registerForActivityResult(ActivityResultContracts.OpenDocument()) { uri ->
        try
        {
            val stream = contentResolver.openInputStream(uri!!)
            val size = stream!!.available()
            val buffer = ByteArray(size)
            stream.read(buffer)
            stream.close()
            initDeviceKit(String(buffer))
        } catch (ex: Exception) {
            ex.printStackTrace()

            if (!MedMDeviceKit.isInitialized())
            {
                mErrorSnackbar = Snackbar.make(findViewById(android.R.id.content), R.string.invalid_licence_key, BaseTransientBottomBar.LENGTH_INDEFINITE)
                    .setAction(R.string.menu_pick_license) { pickLicense() }

                mErrorSnackbar?.show()
            }
        }
    }
    private val mPickFile = registerForActivityResult(ActivityResultContracts.OpenDocument()) { uri ->
        if (uri == null)
            return@registerForActivityResult

        val stream = contentResolver?.openInputStream(uri)
        val size = stream!!.available()
        val buffer = ByteArray(size)
        stream.read(buffer)
        stream.close()
        processPostponedDevicesFile(String(buffer))
    }

    private val deviceList: Array<IDeviceDescription?>
        get()
        {
            try
            {
                return MedMDeviceKit.getDeviceManager().devicesList
            } catch (_: Exception)
            {
            }
            return arrayOfNulls(0)
        }

    override fun onCreate(savedInstanceState: Bundle?)
    {
        super.onCreate(savedInstanceState)
        setContentView(R.layout.main)

        val bottomNav = findViewById<BottomNavigationView>(R.id.bottom_navigation)
        supportFragmentManager.beginTransaction().replace(R.id.fragment_container, mDataCollectionFragment).commit()
        supportActionBar?.title = getString(R.string.title_readings)

        bottomNav.setOnItemSelectedListener {
            lateinit var selectedFragment : ListFragment

            when (it.itemId) {
                R.id.menu_receive_data -> {
                    selectedFragment = mDataCollectionFragment
                    supportActionBar?.title = getString(R.string.title_readings)
                }
                R.id.menu_devices_list -> {
                    selectedFragment = mAddedDevicesFragment
                    supportActionBar?.title = getString(R.string.title_added_devices)
                }
                R.id.menu_add_device -> {
                    selectedFragment = mDeviceScanFragment
                    supportActionBar?.title = getString(R.string.title_scan)
                }
                else -> mDataCollectionFragment
            }

            invalidateOptionsMenu()
            supportFragmentManager.beginTransaction().replace(R.id.fragment_container, selectedFragment).commit()
            true
        }

        val sharedPreferences = getSharedPreferences(packageName, Context.MODE_PRIVATE)
        initDeviceKit(sharedPreferences.getString(LICENSE_SHARED_PREF_ENTRY_NAME, DEVICE_KIT_LICENSE))

        parseSelectedDevices(sharedPreferences.getString(SELECTED_DEVICES_SHARED_PREF_ENTRY_NAME, "[]"))
    }

    private fun initDeviceKit(license: String?)
    {
        var errorResCode = 0
        try
        {
            MedMDeviceKit.init(application, license)
            Toast.makeText(this, "MedMDeviceKit version: " + MedMDeviceKit.getVersion(), Toast.LENGTH_LONG).show()

            val editor = getSharedPreferences(packageName, Context.MODE_PRIVATE).edit()
            editor.putString(LICENSE_SHARED_PREF_ENTRY_NAME, license)
            editor.apply()
            invalidateMenu()
            mErrorSnackbar?.dismiss()
        } catch (ex: InvalidLicenseKeyException)
        {
            ex.printStackTrace()
            errorResCode = R.string.invalid_licence_key
        } catch (ex: BuildExpiredException)
        {
            ex.printStackTrace()
            errorResCode = R.string.build_expired
        } catch (ex: LicenseExpiredException)
        {
            ex.printStackTrace()
            errorResCode = R.string.build_expired
        }

        if (!MedMDeviceKit.isInitialized())
        {
            mErrorSnackbar = Snackbar.make(findViewById(android.R.id.content), errorResCode, BaseTransientBottomBar.LENGTH_INDEFINITE)
                .setAction(R.string.menu_pick_license) { pickLicense() }

            mErrorSnackbar?.show()
        }
    }

    override fun onRestoreInstanceState(state: Bundle)
    {
        super.onRestoreInstanceState(state)
        mSelectedDevices = state.getBooleanArray(SELECTED_DEVICES_SHARED_PREF_ENTRY_NAME) ?: BooleanArray(0)
        updateFilter()
    }

    override fun onSaveInstanceState(outState: Bundle)
    {
        super.onSaveInstanceState(outState)
        outState.putBooleanArray(SELECTED_DEVICES_SHARED_PREF_ENTRY_NAME, mSelectedDevices)
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean
    {
        if (!MedMDeviceKit.isInitialized())
        {
            menuInflater.inflate(R.menu.not_initialized, menu)
            return true
        }

        val bottomNav = findViewById<BottomNavigationView>(R.id.bottom_navigation)
        val menuId = when (bottomNav.selectedItemId)
        {
            R.id.menu_receive_data -> R.menu.readings
            R.id.menu_devices_list -> R.menu.added_devices
            R.id.menu_add_device -> R.menu.scan
            else -> R.menu.not_initialized
        }
        menuInflater.inflate(menuId, menu)

        if (bottomNav.selectedItemId == R.id.menu_receive_data)
        {
            menu.findItem(R.id.menu_collect_stop)?.isVisible = mCollectorStopToken != null
            menu.findItem(R.id.menu_start)?.isVisible = mCollectorStopToken == null
        }
        else if (bottomNav.selectedItemId == R.id.menu_add_device)
        {
            menu.findItem(R.id.menu_scan_stop)?.isVisible = mScanStopToken != null
            menu.findItem(R.id.menu_scan)?.isVisible = mScanStopToken == null
        }

        return true
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean
    {
        val cont = findViewById<FragmentContainerView>(R.id.fragment_container)
        val progress = findViewById<LinearProgressIndicator>(R.id.progress)
        when (item.itemId)
        {
            R.id.menu_devices_filter -> showDeviceFilter()
            R.id.menu_config -> showConfig()
            R.id.menu_pick_license -> pickLicense()
            R.id.menu_user -> showUserSetup()
            R.id.menu_postponed_add -> mPickFile.launch(arrayOf("*/*"))
            R.id.menu_start -> {
                mCollectorStopToken = cont.getFragment<DataCollectionFragment?>()?.startCollectData()
                progress.isVisible = true
            }
            R.id.menu_scan -> {
                mScanStopToken = cont.getFragment<DeviceScanFragment?>()?.startScan()
                progress.isVisible = true
            }
            R.id.menu_scan_stop -> {
                mScanStopToken?.stopScan()
                mScanStopToken = null
                progress.isVisible = mCollectorStopToken != null
            }
                R.id.menu_collect_stop -> {
                mCollectorStopToken?.stopCollect()
                mCollectorStopToken = null
                progress.isVisible = mScanStopToken != null
            }
        }

        invalidateOptionsMenu()
        return true
    }

    private fun showDeviceFilter()
    {
        val builder = MaterialAlertDialogBuilder(this)
        builder.setIcon(R.drawable.ic_launcher)
        builder.setTitle("Select device:")

        val devicesStrings = arrayOfNulls<CharSequence>(deviceList.size + 1)
        devicesStrings[0] = "All devices"
        for (i in deviceList.indices)
            devicesStrings[i + 1] = deviceList[i]!!.modelName + " (" + deviceList[i]!!.address + ")"

        if (mSelectedDevices.size != devicesStrings.size)
        {
            mSelectedDevices = BooleanArray(devicesStrings.size)
            Arrays.fill(mSelectedDevices, true)
        }

        builder.setPositiveButton("Ok") { dialog, _ ->
            dialog.dismiss()
            updateFilter()
            saveSelectedDevices()
        }

        builder.setMultiChoiceItems(devicesStrings, mSelectedDevices) { dialogInterface, index, isSelected ->
            val list = (dialogInterface as AlertDialog).listView
            if (index == 0)
            {
                for (i in 0 until list.count) {
                    list.setItemChecked(i, isSelected)
                }
                Arrays.fill(mSelectedDevices, isSelected)
            } else
            {
                mSelectedDevices[index] = isSelected
                if (!isSelected)
                {
                    mSelectedDevices[0] = false
                    list.setItemChecked(0, false)
                }
            }
        }
        builder.show()
    }

    private fun showConfig()
    {
        val config = MedMDeviceKit.getConfig()
        val builder = MaterialAlertDialogBuilder(this@MainActivity)
        builder.setView(R.layout.config)
        builder.setTitle(getString(R.string.menu_config))

        builder.setPositiveButton("Ok") { dialog, _ ->
            val isMetric = (dialog as AlertDialog).findViewById<RadioGroup>(R.id.units_config_group)!!.checkedRadioButtonId == R.id.units_metric
            val isMillimoles = dialog.findViewById<RadioGroup>(R.id.glucose_units_config_group)!!.checkedRadioButtonId == R.id.units_millimoles
            val convertReadings = dialog.findViewById<SwitchMaterial>(R.id.convert_readings_switch)!!.isChecked
            val controlBluetooth = dialog.findViewById<SwitchMaterial>(R.id.control_bluetooth_state)!!.isChecked
            val enableLogs = dialog.findViewById<SwitchMaterial>(R.id.enable_logs_switch)!!.isChecked
            config.units = if (isMetric) MedMDeviceKitConfig.Units.Metric else MedMDeviceKitConfig.Units.Imperial
            config.glucoseUnits = if (isMillimoles) MedMDeviceKitConfig.GlucoseUnits.Millimoles else MedMDeviceKitConfig.GlucoseUnits.MilligramsPerDecilitre
            config.convertReadings = convertReadings
            config.controlBluetoothState = controlBluetooth
            MedMDeviceKit.setConfig(config)
            MedMDeviceKit.enableLogs(enableLogs)
            dialog.dismiss()
        }
        val dialog = builder.show()
        dialog.findViewById<RadioGroup>(R.id.units_config_group)!!.check(if (config.units == MedMDeviceKitConfig.Units.Metric) R.id.units_metric else R.id.units_imperial)
        dialog.findViewById<RadioGroup>(R.id.glucose_units_config_group)!!.check(if (config.glucoseUnits == MedMDeviceKitConfig.GlucoseUnits.Millimoles) R.id.units_millimoles else R.id.units_milligrams_decilitre)
        dialog.findViewById<SwitchMaterial>(R.id.convert_readings_switch)!!.isChecked = config.convertReadings
        dialog.findViewById<SwitchMaterial>(R.id.control_bluetooth_state)!!.isChecked = config.controlBluetoothState
        dialog.findViewById<SwitchMaterial>(R.id.enable_logs_switch)!!.isChecked = MedMDeviceKit.logsEnabled()
    }

    private fun showUserSetup()
    {
        val config = MedMDeviceKit.getConfig()
        val builder = MaterialAlertDialogBuilder(this@MainActivity)
        builder.setView(R.layout.user)
        builder.setTitle(getString(R.string.menu_user))

        val items = listOf(getString(R.string.female), getString(R.string.male), getString(R.string.unknown), getString(R.string.other))
        val adapter = ArrayAdapter(this, R.layout.listitem_sex, items)

        builder.setPositiveButton("Ok") { dialog, _ ->
            val birthdayEditText = (dialog as AlertDialog).findViewById<TextInputEditText>(R.id.birthday)!!
            val calendar = Calendar.getInstance()
            calendar.time = SimpleDateFormat("dd.MM.yyyy", Locale.ENGLISH).parse(birthdayEditText.text.toString())!!
            config.userBirthday = calendar
            config.userHeight = dialog.findViewById<TextInputEditText>(R.id.height)?.text.toString().toInt()
            config.userWeight = dialog.findViewById<TextInputEditText>(R.id.weight)?.text.toString().toDouble()
            config.userDailyStepsGoal = dialog.findViewById<TextInputEditText>(R.id.steps_goal)?.text.toString().toInt()
            config.userSex = MedMDeviceKitConfig.Sex.values()[items.indexOf(dialog.findViewById<AutoCompleteTextView>(R.id.sex)!!.text.toString()) ]

            MedMDeviceKit.setConfig(config)
            dialog.dismiss()
        }
        val dialog = builder.show()
        val birthdayEditText = dialog.findViewById<TextInputEditText>(R.id.birthday)!!

        if (config.userBirthday != null)
            birthdayEditText.setText(SimpleDateFormat("dd.MM.yyyy", Locale.ENGLISH).format(config.userBirthday.time))

        birthdayEditText.setOnClickListener {
            val selectionDate = config.userBirthday ?: Calendar.getInstance()
            selectionDate.timeZone = TimeZone.getTimeZone("GMT")
            val datePicker =
                MaterialDatePicker.Builder.datePicker()
                    .setTitleText(R.string.birthday)
                    .setSelection(selectionDate.timeInMillis)
                    .build()
            datePicker.addOnPositiveButtonClickListener {
                val calendar = Calendar.getInstance()
                calendar.timeInMillis = it
                birthdayEditText.setText(SimpleDateFormat("dd.MM.yyyy", Locale.ENGLISH).format(calendar.time))
            }
            datePicker.show(supportFragmentManager, "tag")
        }


        val autoCompleteTextView = dialog.findViewById<AutoCompleteTextView>(R.id.sex)!!
        autoCompleteTextView.setAdapter(adapter)
        autoCompleteTextView.setText(items[config.userSex.ordinal], false)

        dialog.findViewById<TextInputEditText>(R.id.height)?.setText(config.userHeight.toString())
        dialog.findViewById<TextInputEditText>(R.id.weight)?.setText(config.userWeight.toString())
        dialog.findViewById<TextInputEditText>(R.id.steps_goal)?.setText(config.userDailyStepsGoal.toString())
    }

    private fun updateFilter()
    {
        val devices = MedMDeviceKit.getDeviceManager().devicesList

        if (devices.isEmpty())
        {
            mSelectedDevices = BooleanArray(1)
            mSelectedDevices[0] = true
            return
        }

        if (devices.size + 1 != mSelectedDevices.size)
        {
            mSelectedDevices = BooleanArray(devices.size + 1)
            Arrays.fill(mSelectedDevices, true)
            MedMDeviceKit.getCollector().deviceFilter = devices
            return
        }

        if (mSelectedDevices[0])
        {
            MedMDeviceKit.getCollector().deviceFilter = devices
            return
        }

        val array = ArrayList<IDeviceDescription>()
        for (i in 1 until mSelectedDevices.size)
            if (mSelectedDevices[i]) array.add(devices[i - 1])

        MedMDeviceKit.getCollector().deviceFilter = array.toTypedArray()
    }

    private fun saveSelectedDevices()
    {
        val editor = getSharedPreferences(packageName, Context.MODE_PRIVATE).edit()
        editor.putString(SELECTED_DEVICES_SHARED_PREF_ENTRY_NAME, mSelectedDevices.contentToString())
        editor.apply()
    }

    private fun pickLicense()
    {
        mPickLicense.launch(arrayOf("*/*"))
    }

    private fun parseSelectedDevices(source: String?)
    {
        if (source == null)
        {
            mSelectedDevices = BooleanArray(0)
            return
        }

        val parts = source.substring(1, source.length - 1).split(",").toTypedArray()
        mSelectedDevices = BooleanArray(parts.size)
        for (i in mSelectedDevices.indices)
            mSelectedDevices[i] = java.lang.Boolean.parseBoolean(parts[i].trim { it <= ' ' })
    }

    private fun processPostponedDevicesFile(json: String)
    {
        try {
            val array = JSONArray(json)
            for (i in 0 until array.length())
            {
                val deviceJson = array.getJSONObject(i)

                val sku = deviceJson.getInt("sku")
                val address = deviceJson.optString("address")
                val pin = deviceJson.optString("pin")

                try {
                    if (address.isNotEmpty())
                        MedMDeviceKit.getDeviceManager().addDevicePostponed(sku, PostponedAddFilter.fromAddress(address), pin)
                    else
                        MedMDeviceKit.getDeviceManager().addDevicePostponed(sku, PostponedAddFilter.fromSerial(deviceJson.getString("serial")), pin)
                } catch (e: FailedToAddPostponedDeviceException) {
                    Toast.makeText(this, "Failed to add device", Toast.LENGTH_SHORT).show()
                    return
                } catch (e: FeatureUnsupportedException) {
                    Toast.makeText(this, "Feature unsupported", Toast.LENGTH_SHORT).show()
                    return
                }
            }
        } catch (e: Exception) {
            Toast.makeText(this, "Wrong format", Toast.LENGTH_SHORT).show()
            return
        }
    }
}